#Jacob's Emulation of Low Level Operations
#JELLO System
import time
import sys

help = """Useage: jello [flags] [options] file_name
Flags:
  -v: Displays version
  -h: Displays this help section
Options:
  -dbl: Sets the debug level, can be 0 to 2
  -t: Sets internal clock rate in hertz
"""

version       = "0.6.4 (Menma) by Jacob Allen"
version_date  = "02 October 2015"
print_level   = 0
timer         = 0

args = sys.argv[1:]

file_name = ""

if len(args) < 1:
  print "Command requires arguments."
  sys.exit(0)

i = 0
options = {}
flags = ["-v","-h"]
args_to_process = args
while len(args_to_process) > 0:
  item = args_to_process.pop(0)
  if item[0] == "-":
    if item in flags:
      options[item] = True
    else:
      if len(args_to_process) > 0:
        options[item] = args_to_process.pop(0)
      else:
        print "Options requires value."
        sys.exit(0)
  else:
    if file_name == "":
      file_name = item
    else:
      print "Only one file name is required."
      sys.exit(0)

if "-h" in options.keys():
  print "JELLO V"+version+" - "+version_date
  print help
  sys.exit()

if "-v" in options.keys():
  print "JELLO V"+version+" - "+version_date
  sys.exit()

if "-dbl" in options.keys():
  if options["-dbl"].isdigit():
    print_level = int(options["-dbl"])
  else:
    print "Debug level must be an integer."
    sys.exit(0)

if "-t" in options.keys():
  if options["-t"].replace("hz","").replace(".","").isdigit():
    timer = float(options["-t"].replace("hz",""))
  else:
    print "Clock speed must be in hertz."
    sys.exit(0)

if file_name == "":
  print "A file name is required."
  sys.exit(0)

print "JELLO V"+version

print 'Reading file "'+file_name+'"...',
fp = open(file_name)
code = fp.read()
fp.close
print "Done."

class intst:
  reg    = 0
  mem    = {}
  input  = 0
  output = 0

def interpret_value(val,retval=False):
  if val.isdigit():
    return int(val)
  elif val == "REG":
    return intst.reg
  elif val == "INP":
    return intst.input
  elif val[:4] == "MEM(" and val[-1] == ")":
    loc = val.replace("MEM(","").replace(")","")
    if interpret_value(loc) in intst.mem.keys():
      return intst.mem[interpret_value(loc)]
  return val if retval else 0

def find_next_end(p):
  while p < len(lines):
    if lines[p].strip() == "END": return p
    p += 1
  return 0

def find_label(name):
  p = 0
  while p < len(lines):
    if lines[p].strip() == str(name): return p
    p += 1
  return 0

def do_error(type="Error"):
  print type.upper()+" AT "+str(prev_pointer+1)
  print "  ->", line.strip()
  print "     @REG:", intst.reg
  print "     @MEM:", intst.mem
  print "     @INP:", intst.input
  print "     @OUT:", intst.output
  print "Program failed to complete due to an error."
  sys.exit()

def cleanup():
  global pointer
  if intst.reg > 255:
    do_error("Register Overflow Error")
  if intst.input > 255:
    do_error("Input Overflow Error")
  if intst.output > 255:
    do_error("Output Overflow Error")
  for item in intst.mem.values():
    if item > 255:
      do_error("Memory Overflow Error")
  if pointer < 0 or pointer > len(lines):
    do_error("Line Pointer Error")
    

def process_command(command, vals):
  global pointer
  if command == "ADD" and len(vals) == 1:
    intst.reg += interpret_value(vals[0])
    return 1
  elif command == "SUB" and len(vals) == 1:
    intst.reg -= interpret_value(vals[0])
    return 1
  elif command == "MUL" and len(vals) == 1:
    intst.reg *= interpret_value(vals[0])
    return 1
  elif command == "DIV" and len(vals) == 1:
    intst.reg /= interpret_value(vals[0])
    return 1
  elif command == "SETMEM" and len(vals) == 2:
    intst.mem[interpret_value(vals[0])] = interpret_value(vals[1])
    return 1
  elif command == "SETREG" and len(vals) == 1:
    intst.reg = interpret_value(vals[0])
    return 1
  elif command == "SETOUT" and len(vals) == 1:
    intst.output = interpret_value(vals[0])
    print "  ->", "OUT:", intst.output
    return 1
  elif command == "GETINP" and len(vals) == 0:
    print "     GIVE INPUT:",
    intst.input = interpret_value(raw_input().strip())
    return 1
  elif command == "END" and len(vals) == 0:
    return 1
  elif command == "GOTO" and len(vals) == 1:
    if vals[0][0] == ":":
      return find_label(interpret_value(vals[0], True)) - pointer
    elif vals[0][0] == "+" or vals[0][0] == "-":
      return int(vals[0][0]+str(interpret_value(vals[0][1:])))
    elif type(interpret_value(vals[0])) is int:
      return interpret_value(vals[0]) - 1 - pointer
    return 1
  elif command == "IF" and len(vals) == 2:
    if vals[0] == "EQ":
      if interpret_value(vals[1]) == intst.reg:
        return 1
      else:
        return find_next_end(pointer) - pointer
    elif vals[0] == "GT":
      if interpret_value(vals[1]) < intst.reg:
        return 1
      else:
        return find_next_end(pointer) - pointer
    elif vals[0] == "LT":
      if interpret_value(vals[1]) > intst.reg:
        return 1
      else:
        return find_next_end(pointer) - pointer
    return 1
  elif command == "IF" and len(vals) == 1:
    if vals[0] == "ZERO":
      if intst.reg == 0:
        return 1
      else:
        return find_next_end(pointer) - pointer
    return 1
  do_error("Unsupported Command Error")

lines = code.replace("\r\n","\n").split("\n")

print "Processing Commands..."
pointer = 0
prev_pointer = 0
while pointer < len(lines):
  line = lines[pointer]
  line = line.strip()
  
  if line == "" or line[0] == "#" or line[0] == ":":
    pointer += 1
    continue
  
  parts = line.split()
  command = parts.pop(0)
  values = parts
  if print_level > 0: print "  ->", line.strip(), "AT", pointer+1
  prev_pointer = pointer
  p_inc = process_command(command, values)
  if p_inc == None: do_error("Line Pointer Error")
  pointer += p_inc
  cleanup()
  if print_level > 1:
    print "     @REG:", intst.reg
    print "     @MEM:", intst.mem
    print "     @INP:", intst.input
    print "     @OUT:", intst.output
  if timer > 0: time.sleep(1.0/timer)
print "Done."